package com.bytes.articles.java.generics;

import java.nio.CharBuffer;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

/**
 * A small class for demonstrating the use of wildcards in Java Generics
 * @author Nepomuk
 *
 * @param <T>
 */
public class Wildcards<T> {
	
	/**
	 * An internal List that can be set with the setter.
	 */
	private List<T> list;
	
	/**
	 * The above mentioned setter.
	 * @param list
	 */
	public void setList(List<T> list) {
		this.list = list;
	}
	
	/**
	 * Checks, whether the given list has more elements than the internal one
	 * @param otherList
	 * @return <b>true</b> if the internal list is bigger, <b>false</b> otherwise
	 */
	public boolean hasMoreElementsThan(List<?> otherList) {
		return list.size() > otherList.size();
	}
	
	/**
	 * Checks whether the size of the internal list can be approximated by one of the entries in the given list.
	 * @param otherList
	 * @return <b>true</b> if <i>Number#intValue()</i> of one of the entries equals the size of the internal list.
	 */
	public boolean closeToSize(List<? extends Number> otherList) {
		for(Number item : otherList) {
			if(item.intValue() == list.size())
				return true;
		}
		return false;
	}
	
	/**
	 * The same as closeToSize(), but the type is available.
	 * @param otherList
	 * @return <b>true</b> if <i>Number#intValue()</i> of one of the entries equals the size of the internal list.
	 */
	public <S extends Number> boolean closeToSize2(List<S> otherList) {
		for(int i=0; i<otherList.size(); i++) {
			S item = otherList.get(i);
			if(item.intValue() == list.size())
				return true;
		}
		return false;
	}
	
	/**
	 * Unite two lists of forms.
	 * @param listA
	 * @param listB
	 */
	public void unite(List<? extends Circle> listA, List<? super Ellipse> listB) {
		for(Circle item : listA) {
			listB.add(item);
		}
		for(Object obj : listB) {
			Ellipse ellipse = (Ellipse) obj;
			System.out.println("An ellipse with the area " + ellipse.getArea() + " was found.");
		}
	}
	
	/**
	 * Inspects the elements in a list with "instanceof"
	 * @param list
	 */
	public void inspect(List<?> list) {
		for(int i=0; i<list.size(); i++) {
			if (list.get(i) instanceof Object)
				System.out.println(i + ". Look, an Object!");
			if (list.get(i) instanceof String)
				System.out.println("And it's a String too!");
			else if(list.get(i) instanceof Integer)
				System.out.println("And it's an Integer too!");
			else if(list.get(i) instanceof Double)
				System.out.println("And it's a Double too!");
			else if(list.get(i) instanceof Circle)
				System.out.println("And it's a Circle too!");
			else if(list.get(i) instanceof Ellipse)
				System.out.println("And it's an Ellipse too!");
		}
	}
	
	/**
	 * A main function, so that this class is runnable
	 * @param args
	 */
	public static void main(String[] args) {
		// Create a new ArrayList of Strings and fill it with "Hello", "World" and "!"
		ArrayList<String> list1 = new ArrayList<String>();
		list1.add("Hello");
		list1.add("World");
		list1.add("!");
		// Create a new Wildcard<String> object and set list1 as the internal list 
		Wildcards<String> w1 = new Wildcards<String>();
		w1.setList(list1);
		// Create a new LinkedList of Integers and fill it with the Numbers 0 through 9
		ArrayList<Integer> list2 = new ArrayList<Integer>();
		for(int i=0; i<10; i++) {
			list2.add(new Integer(i));
		}
		// Use the <i>hasMoreElementsThan</i> function
		if(w1.hasMoreElementsThan(list2)) {
			System.out.println("The internal list is bigger than the one it was compared to");
		} else {
			System.out.println("The internal list is not bigger than the one it was compared to");
		}
		// A list of Doubles
		ArrayList<Double> list3 = new ArrayList<Double>();
		list3.add(new Double("1.2"));
		list3.add(new Double("2.3"));
		list3.add(new Double("3.4"));
		list3.add(new Double("4.5"));
		// Check, whether the size of list1 can be approximated through the values in list3
		if(w1.closeToSize(list3)) {
			System.out.println("The size of list1 can be approximated through a value in list3");
		} else {
			System.out.println("The size of list1 can not be approximated through a value in list3");
		}
		// Let's create a Wildcard object for list2 and try again
		Wildcards<Integer> w2 = new Wildcards<Integer>();
		w2.setList(list2);
		if(w2.closeToSize(list3)) {
			System.out.println("The size of list2 can be approximated through a value in list3");
		} else {
			System.out.println("The size of list2 can not be approximated through a value in list3");
		}
		// Now let's test our "unite" function:
		ArrayList<Circle> list4 = new ArrayList<Circle>();
		list4.add(new Circle(1));
		list4.add(new Circle(3));
		list4.add(new Circle(5));
		ArrayList<Ellipse> list5 = new ArrayList<Ellipse>();
		list5.add(new Ellipse(1,2));
		list5.add(new Ellipse(3,4));
		w1.unite(list4, list5);
		// Inspect our lists
		System.out.println("\n   Inspect list1");
		w1.inspect(list1);
		System.out.println("\n   Inspect list2");
		w1.inspect(list2);
		System.out.println("\n   Inspect list3");
		w1.inspect(list3);
		System.out.println("\n   Inspect list4");
		w1.inspect(list4);
		System.out.println("\n   Inspect list5");
		w1.inspect(list5);
	}
}

abstract class Form {
	protected double area;
	
	public double getArea() {
		return area;
	}
}

class Ellipse extends Form {
	public Ellipse(int rx, int ry) {
		super();
		area = Math.PI * rx * ry;
	}
}

class Circle extends Ellipse {
	public Circle(int r) {
		super(r,r);
	}
}

class Rectangle extends Form {
	public Rectangle(int x, int y) {
		area = x * y;
	}
}

class Square extends Rectangle {
	public Square(int x) {
		super(x,x);
	}
}